// Protocol Buffers for Swift
//
// Copyright 2014 Alexey Khohklov(AlexeyXo).
// Copyright 2008 Google Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "swift_generator.h"

#include <google/protobuf/io/printer.h>
#include <google/protobuf/io/zero_copy_stream.h>
#include <google/protobuf/descriptor.pb.h>
#include <google/protobuf/stubs/strutil.h>

#include "swift_file.h"
#include "swift_helpers.h"

namespace google { namespace protobuf { namespace compiler { namespace swift {
    SwiftGenerator::SwiftGenerator() {
    }


    SwiftGenerator::~SwiftGenerator() {
    }


    bool SwiftGenerator::Generate(const FileDescriptor* file,
                                  const string& parameter,
                                  GeneratorContext* generator_context,
                                  string* error) const {
        
        

        static map<string, string> packages;
        
        vector< pair<string, string> > options;
        ParseGeneratorParameter(parameter, &options);

        for (int i = 0; i < options.size(); i++) {
            *error = "Unknown generator option: " + options[i].first;
            return false;
        }

        FileGenerator file_generator(file);

        string filepath = FilePath(file);
        {
           
            string package_name;
           
            if (file->package() != "") {
                package_name = UnderscoresToCapitalizedCamelCase(file->package()) + "_";
            }
            
            vector<string> tokens = FullNameSplit(file->package());
            
            scoped_ptr<io::ZeroCopyOutputStream> output(generator_context->Open(package_name + filepath + ".pb.swift"));
            io::Printer printer(output.get(), '$');
            
            
            printer.Print("// Generated by the protocol buffer compiler.  DO NOT EDIT!\n\n"
                          "import Foundation\n");
            if (isCompileForFramework(file)) {
                printer.Print("import ProtocolBuffers\n\n");
            }
            printer.Print("\n");
            
            bool extensionGenerated = false;
            
            vector<string> structsGenerated;
            
            if (!IsBootstrapFile(file)) {
                for (int i = 0; i < tokens.size(); i++) {
                    
                    if (packages.find(tokens[i]) == packages.end() && !IsBootstrapPackage(tokens[i])) {
                        (packages[tokens[i]] = "generate");
                        
                        if (extensionGenerated) {
                            printer.Print("{");
                            structsGenerated.push_back(tokens[i]);
                        }
                        printer.Print("$acontrol$ struct $package$ {",
                                      "acontrol", GetAccessControlType(file),
                                      "package", tokens[i]);
                        structsGenerated.push_back(tokens[i]);
                        extensionGenerated = false;
                    }
                    else
                    {
                        if (i == 0) {
                            printer.Print("$acontrol$ extension $package$",
                                          "acontrol", GetAccessControlType(file),
                                          "package", tokens[i]);
                        }
                        else
                        {
                            printer.Print(".$package$",
                                          "package", tokens[i]);
                        }
                        if (i == tokens.size() - 1) {
                            printer.Print(" {}");
                        }
                        extensionGenerated = true;
                        
                    }
                }
                
                for (int i = 0; i < structsGenerated.size(); i++) {
                    printer.Print("}");
                }
                printer.Print("\n");
            }
            else
            {
                for (int i = 0; i < tokens.size(); i++) {
                    
                    if (packages.find(tokens[i]) == packages.end() && IsBootstrapPackage(tokens[i])) {
                        (packages[tokens[i]] = "generate");
                        
                        if (extensionGenerated) {
                            printer.Print("{");
                            structsGenerated.push_back(tokens[i]);
                        }
                        printer.Print("$acontrol$ struct $package$ {",
                                      "acontrol", GetAccessControlType(file),
                                      "package", tokens[i]);
                        structsGenerated.push_back(tokens[i]);
                        extensionGenerated = false;
                    }
                    else
                    {
                        if (i == 0) {
                            printer.Print("$acontrol$ extension $package$",
                                          "acontrol", GetAccessControlType(file),
                                          "package", tokens[i]);
                        }
                        else
                        {
                            printer.Print(".$package$",
                                          "package", tokens[i]);
                        }
                        extensionGenerated = true;
                        if (i == tokens.size() - 1) {
                            printer.Print(" {}");
                        }
                        
                    }
                }
                
                for (int i = 0; i < structsGenerated.size(); i++) {
                    printer.Print("}");
                }
                printer.Print("\n");

            }
            
            file_generator.GenerateSource(&printer);
        }

        return true;
    }

}  // namespace swift
}  // namespace compiler
}  // namespace protobuf
}  // namespace google
